/*
  WiFiTelnetToSerial - Example Transparent UART to Telnet Server for ESP32
  Extended with WifiManager by RFXCOM

  Copyright (c) 2017 Hristo Gochkov. All rights reserved.
  This file is part of the ESP32 WiFi library for Arduino environment.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

  Version 0.0 12 Nov 2022
    Initial RFXCOM version
  Version 1.0 12 dec 2022
    Wipe EEPROM check at start
    Reset microcontroller after Wifi connection
  Version 1.1 12 feb 2023
    delay 100msec after restore button pressed
*/

#include <WiFi.h>
#include <WiFiMulti.h>
#include "EEPROM.h"
#include "HTML.h"
#include "WebServer.h"
#include "WifiFunc.h"

WiFiMulti wifiMulti;

//how many clients should be able to telnet to this ESP32
#define MAX_SRV_CLIENTS 2

WiFiServer server(10001);
WiFiClient serverClients[MAX_SRV_CLIENTS];

void setup() {
  Serial.begin(115200);
  Serial.println("\nConnecting");

  pinMode(18, INPUT_PULLUP);  //for resetting WiFi credentials
  
/*
EEPROM storage needed / eeprom address:
ssid    31    0
passw   31    31
ip      16    62
subnet  16    78
gateway 16    94
TOTAL         110
*/
  EEPROM.begin(110);

  if (digitalRead(18) == LOW) {
    delay(100);  
    if (digitalRead(18) == LOW) {    
      Serial.println("Wiping WiFi credentials from memory...");
      wipeEEPROM();
    }
  }
  
  Serial.println("Checking WIFI credentials");

  String s = EEPROM.readString(0);
  String p = EEPROM.readString(31);
  String i = EEPROM.readString(62);
  String n = EEPROM.readString(78);
  String g = EEPROM.readString(94);

  //convert strings to char array
  char ssid[s.length() + 1];
  s.toCharArray(ssid, s.length() + 1);

  char passw[p.length() + 1];
  p.toCharArray(passw, p.length() + 1);

//#if DEBUG
  Serial.print("Found credentials: ");
  Serial.print(s);
  Serial.print("/");
  Serial.print(p);
  Serial.print("/");
  Serial.print(i);
  Serial.print("/");
  Serial.print(n);
  Serial.print("/");
  Serial.println(g);
  delay(5000);
//#endif

  if (i.length() > 0 && n.length() > 0 && g.length() > 0) {

    IPAddress staticIP;
    staticIP.fromString(i);

    IPAddress gateway;
    gateway.fromString(g);

    IPAddress subnet;
    subnet.fromString(n);

    // Configures static IP address
    if (!WiFi.config(staticIP, gateway, subnet)) {
      Serial.println("STA IP failed to configure");
    }
  } else {
    Serial.println("DHCP used");
  }

  if (s.length() > 0 && p.length() > 0) {
    wifiMulti.addAP(ssid, passw);
    //wifiMulti.addAP("ssid_from_AP_2", "your_password_for_AP_2");
    //wifiMulti.addAP("ssid_from_AP_3", "your_password_for_AP_3");
  } else {
    Serial.println("No WIFI credentials stored in memory. Loading form...");
    while (loadWIFICredsForm())
      ;
  }

  Serial.println("Connecting Wifi ");
  for (int loops = 10; loops > 0; loops--) {
    if (wifiMulti.run() == WL_CONNECTED) {
      Serial.println("");
      Serial.print("WiFi connected ");
      Serial.print("IP address: ");
      Serial.println(WiFi.localIP());
      break;
    } else {
      Serial.println(loops);
      delay(1000);
    }
  }
  if (wifiMulti.run() != WL_CONNECTED) {
    Serial.println("WiFi connect failed");
    delay(1000);
    ESP.restart();
  }

  
  //start UART and the server
  Serial2.begin(38400);
  server.begin();
  server.setNoDelay(true);

  pinMode(19, OUTPUT);
  digitalWrite(19, LOW);  //reset microcontroller
  delay(100);             //delay 0.1sec
  pinMode(19, INPUT);     //high impedance
  
#if DEBUG
  Serial.print("Ready! Use 'telnet ");
  Serial.print(WiFi.localIP());
  Serial.println(" 10001' to connect");
#endif  
}



//**************************** Telnet server ************************************
void loop() {
  uint8_t i;

  if (digitalRead(18) == LOW) {
    delay(100);  
    if (digitalRead(18) == LOW) {    
      Serial.println("Wiping WiFi credentials from memory...");
      wipeEEPROM();
      while (loadWIFICredsForm());
    }    
    Serial.println("Wiping WiFi credentials from memory...");
    wipeEEPROM();

  }

  if (wifiMulti.run() == WL_CONNECTED) {
    //check if there are any new clients
    if (server.hasClient()) {
      for (i = 0; i < MAX_SRV_CLIENTS; i++) {
        //find free/disconnected spot
        if (!serverClients[i] || !serverClients[i].connected()) {
          if (serverClients[i]) serverClients[i].stop();
          serverClients[i] = server.available();
          if (!serverClients[i]) Serial.println("available broken");
          Serial.print("New client: ");
          Serial.print(i);
          Serial.print(' ');
          Serial.println(serverClients[i].remoteIP());
          break;
        }
      }
      if (i >= MAX_SRV_CLIENTS) {
        //no free/disconnected spot so reject
        server.available().stop();
      }
    }
    //check clients for data
    for (i = 0; i < MAX_SRV_CLIENTS; i++) {
      if (serverClients[i] && serverClients[i].connected()) {
        if (serverClients[i].available()) {
          //get data from the telnet client and push it to the UART
          while (serverClients[i].available()) Serial2.write(serverClients[i].read());
        }
      } else {
        if (serverClients[i]) {
          serverClients[i].stop();
        }
      }
    }
    //check UART for data
    if (Serial2.available()) {
      size_t len = Serial2.available();
      uint8_t sbuf[len];
      Serial2.readBytes(sbuf, len);
      //push UART data to all connected telnet clients
      for (i = 0; i < MAX_SRV_CLIENTS; i++) {
        if (serverClients[i] && serverClients[i].connected()) {
          serverClients[i].write(sbuf, len);
          delay(1);
        }
      }
    }
  } else {
    Serial.println("WiFi not connected!");
    for (i = 0; i < MAX_SRV_CLIENTS; i++) {
      if (serverClients[i]) serverClients[i].stop();
    }
    delay(1000);
  }
}

void wipeEEPROM() {
  for (int i = 0; i < 110; i++) {
    EEPROM.writeByte(i, 0);
  }
  EEPROM.commit();
}